@@ -13,28 +13,28 @@
 # You should have received a copy of the GNU Affero General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import unicode_literals
 
-import logging
 import datetime
 import importlib
+import logging
 from functools import wraps
-import isodate
 from xml.parsers import expat
 
+import isodate
+import lasso
+from django.conf import settings
 from django.contrib import auth
 from django.template.loader import render_to_string
 from django.urls import reverse
 from django.utils.encoding import force_text
-from django.utils.timezone import make_aware, now, make_naive, is_aware, get_default_timezone
-from django.conf import settings
 from django.utils.six.moves.urllib.parse import urlparse
-import lasso
+from django.utils.timezone import get_default_timezone, is_aware, make_aware, make_naive, now
 
 from . import app_settings
 
 logger = logging.getLogger(__name__)
 
+
 def create_metadata(request):
     entity_id = reverse('mellon_metadata')
     login_url = reverse(app_settings.LOGIN_URL)
@@ -54,13 +54,13 @@ def create_metadata(request):
         'logout_url': request.build_absolute_uri(logout_url),
         'public_keys': public_keys,
         'name_id_formats': name_id_formats,
+        'assertion_consumer_bindings': app_settings.ASSERTION_CONSUMER_BINDINGS,
         'default_assertion_consumer_binding': app_settings.DEFAULT_ASSERTION_CONSUMER_BINDING,
         'organization': app_settings.ORGANIZATION,
         'contact_persons': app_settings.CONTACT_PERSONS,
     }
     if app_settings.METADATA_PUBLISH_DISCOVERY_RESPONSE:
-        ctx['discovery_endpoint_url'] = request.build_absolute_uri(
-            reverse('mellon_login'))
+        ctx['discovery_endpoint_url'] = request.build_absolute_uri(reverse('mellon_login'))
     return render_to_string('mellon/metadata.xml', ctx)
 
 
@@ -81,8 +81,9 @@ def create_server(request):
         else:  # no signature
             private_key = None
             private_key_password = None
-        server = lasso.Server.newFromBuffers(metadata, private_key_content=private_key,
-                                             private_key_password=private_key_password)
+        server = lasso.Server.newFromBuffers(
+            metadata, private_key_content=private_key, private_key_password=private_key_password
+        )
         if app_settings.SIGNATURE_METHOD:
             symbol_name = 'SIGNATURE_METHOD_' + app_settings.SIGNATURE_METHOD.replace('-', '_').upper()
             if hasattr(lasso, symbol_name):
@@ -132,8 +133,7 @@ def get_idp(entity_id):
 def get_idps():
     for adapter in get_adapters():
         if hasattr(adapter, 'get_idps'):
-            for idp in adapter.get_idps():
-                yield idp
+            yield from adapter.get_idps()
 
 
 def flatten_datetime(d):
@@ -145,10 +145,10 @@ def flatten_datetime(d):
 
 
 def iso8601_to_datetime(date_string, default=None):
-    '''Convert a string formatted as an ISO8601 date into a datetime
-       value.
+    """Convert a string formatted as an ISO8601 date into a datetime
+    value.
 
-       This function ignores the sub-second resolution'''
+    This function ignores the sub-second resolution"""
     try:
         dt = isodate.parse_datetime(date_string)
     except Exception:
@@ -170,6 +170,7 @@ def to_list(func):
     @wraps(func)
     def f(*args, **kwargs):
         return list(func(*args, **kwargs))
+
     return f
 
 
@@ -198,9 +199,9 @@ def get_values(saml_attributes, name):
 
 
 def get_setting(idp, name, default=None):
-    '''Get a parameter from an IdP specific configuration or from the main
-       settings.
-    '''
+    """Get a parameter from an IdP specific configuration or from the main
+    settings.
+    """
     return idp.get(name) or getattr(app_settings, name, default)
 
 
@@ -212,14 +213,16 @@ def make_session_dump(lasso_name_id, indexes):
     sp_name_qualifier = lasso_name_id.spNameQualifier and force_text(lasso_name_id.spNameQualifier)
     for index in indexes:
         issuer = index.saml_identifier.issuer
-        session_infos.append({
-            'entity_id': issuer,
-            'session_index': index.session_index,
-            'name_id_content': name_id,
-            'name_id_format': name_id_format,
-            'name_id_name_qualifier': name_qualifier,
-            'name_id_sp_name_qualifier': sp_name_qualifier,
-        })
+        session_infos.append(
+            {
+                'entity_id': issuer,
+                'session_index': index.session_index,
+                'name_id_content': name_id,
+                'name_id_format': name_id_format,
+                'name_id_name_qualifier': name_qualifier,
+                'name_id_sp_name_qualifier': sp_name_qualifier,
+            }
+        )
     session_dump = render_to_string('mellon/session_dump.xml', {'session_infos': session_infos})
     return session_dump
 
@@ -286,6 +289,7 @@ def get_xml_encoding(content):
 
         if encoding:
             xml_encoding = encoding
+
     parser = expat.ParserCreate()
     parser.XmlDeclHandler = xmlDeclHandler
     try:
@@ -301,5 +305,5 @@ def get_local_path(request, url):
     parsed = urlparse(url)
     path = parsed.path
     if request.META.get('SCRIPT_NAME'):
-        path = path[len(request.META['SCRIPT_NAME']):]
+        path = path[len(request.META['SCRIPT_NAME']) :]
     return path
