@@ -13,46 +13,44 @@
 # You should have received a copy of the GNU Affero General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import unicode_literals
 
-from importlib import import_module
 import logging
-import requests
-import lasso
 import uuid
-from requests.exceptions import RequestException
-from xml.sax.saxutils import escape
 import xml.etree.ElementTree as ET
+from contextlib import contextmanager, nullcontext
+from importlib import import_module
+from io import StringIO
+from xml.sax.saxutils import escape
 
 import django.http
-from django.views.generic import View
-from django.http import HttpResponseRedirect, HttpResponse
-from django.contrib import auth
-from django.contrib.auth import get_user_model
+import lasso
+import requests
 from django.conf import settings
-from django.views.decorators.csrf import csrf_exempt
+from django.contrib import auth
+from django.contrib.auth import REDIRECT_FIELD_NAME, get_user_model
+from django.db import transaction
+from django.http import Http404, HttpResponse, HttpResponseForbidden, HttpResponseRedirect
 from django.shortcuts import render, resolve_url
 from django.urls import reverse
-from django.utils.http import urlencode
 from django.utils import six
-from django.utils.encoding import force_text, force_str
-from django.contrib.auth import REDIRECT_FIELD_NAME
-from django.db import transaction
+from django.utils.encoding import force_str, force_text
+from django.utils.http import urlencode
 from django.utils.translation import ugettext as _
+from django.views.decorators.csrf import csrf_exempt
+from django.views.generic import View
+from django.views.generic.base import RedirectView
+from requests.exceptions import RequestException
 
-from . import app_settings, utils, models
-
+from . import app_settings, models, utils
 
 RETRY_LOGIN_COOKIE = 'MELLON_RETRY_LOGIN'
 
 lasso.setFlag('thin-sessions')
 
-if six.PY3:
-    def lasso_decode(x):
-        return x
-else:
-    def lasso_decode(x):
-        return x.decode('utf-8')
+
+def lasso_decode(x):
+    return x
+
 
 EO_NS = 'https://www.entrouvert.com/'
 LOGIN_HINT = '{%s}login-hint' % EO_NS
@@ -63,15 +61,16 @@ User = get_user_model()
 class HttpResponseBadRequest(django.http.HttpResponseBadRequest):
     def __init__(self, *args, **kwargs):
         kwargs['content_type'] = kwargs.get('content_type', 'text/plain')
-        super(HttpResponseBadRequest, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self['X-Content-Type-Options'] = 'nosniff'
 
 
-class LogMixin(object):
+class LogMixin:
     """Initialize a module logger in new objects"""
+
     def __init__(self, *args, **kwargs):
         self.log = logging.getLogger(__name__)
-        super(LogMixin, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
 
 def check_next_url(request, next_url):
@@ -92,7 +91,7 @@ def check_next_url(request, next_url):
     return next_url
 
 
-class ProfileMixin(object):
+class ProfileMixin:
     profile = None
 
     def set_next_url(self, next_url):
@@ -136,6 +135,11 @@ class ProfileMixin(object):
 
 
 class LoginView(ProfileMixin, LogMixin, View):
+    def dispatch(self, request, *args, **kwargs):
+        self.debug_login = request.session.get('mellon_debug_login')
+        with self.capture_logs() if self.debug_login else nullcontext():
+            return super().dispatch(request, *args, **kwargs)
+
     @property
     def template_base(self):
         return self.kwargs.get('template_base', 'base.html')
@@ -159,8 +163,12 @@ class LoginView(ProfileMixin, LogMixin, View):
     def post(self, request, *args, **kwargs):
         '''Assertion consumer'''
         if 'SAMLart' in request.POST:
+            if 'artifact' not in app_settings.ASSERTION_CONSUMER_BINDINGS:
+                raise Http404('artifact binding is not supported')
             return self.continue_sso_artifact(request, lasso.HTTP_METHOD_ARTIFACT_POST)
         if 'SAMLResponse' not in request.POST:
+            if 'post' not in app_settings.ASSERTION_CONSUMER_BINDINGS:
+                raise Http404('post binding is not supported')
             return self.get(request, *args, **kwargs)
         if not utils.is_nonnull(request.POST['SAMLResponse']):
             return HttpResponseBadRequest('SAMLResponse contains a null character')
@@ -173,13 +181,16 @@ class LoginView(ProfileMixin, LogMixin, View):
             login.processAuthnResponseMsg(request.POST['SAMLResponse'])
             login.acceptSso()
         except lasso.ProfileCannotVerifySignatureError:
-            self.log.warning('SAML authentication failed: signature validation failed for %r',
-                             login.remoteProviderId)
+            self.log.warning(
+                'SAML authentication failed: signature validation failed for %r', login.remoteProviderId
+            )
         except lasso.ParamError:
             self.log.exception('lasso param error')
-        except (lasso.LoginStatusNotSuccessError,
-                lasso.ProfileStatusNotSuccessError,
-                lasso.ProfileRequestDeniedError):
+        except (
+            lasso.LoginStatusNotSuccessError,
+            lasso.ProfileStatusNotSuccessError,
+            lasso.ProfileRequestDeniedError,
+        ):
             self.show_message_status_is_not_success(login, 'SAML authentication failed')
         except lasso.Error as e:
             return HttpResponseBadRequest('error processing the authentication response: %r' % e)
@@ -195,15 +206,15 @@ class LoginView(ProfileMixin, LogMixin, View):
         idp = utils.get_idp(login.remoteProviderId)
         if not idp:
             self.log.warning('entity id %r is unknown', login.remoteProviderId)
-            return HttpResponseBadRequest(
-                'entity id %r is unknown' % login.remoteProviderId)
+            return HttpResponseBadRequest('entity id %r is unknown' % login.remoteProviderId)
         error_url = utils.get_setting(idp, 'ERROR_URL')
         error_redirect_after_timeout = utils.get_setting(idp, 'ERROR_REDIRECT_AFTER_TIMEOUT')
         if error_url:
             error_url = resolve_url(error_url)
         next_url = error_url or self.get_next_url(default=resolve_url(settings.LOGIN_REDIRECT_URL))
         return self.render(
-            request, 'mellon/authentication_failed.html',
+            request,
+            'mellon/authentication_failed.html',
             {
                 'debug': settings.DEBUG,
                 'reason': reason,
@@ -212,7 +223,8 @@ class LoginView(ProfileMixin, LogMixin, View):
                 'next_url': next_url,
                 'relaystate': login.msgRelayState,
                 'error_redirect_after_timeout': error_redirect_after_timeout,
-            })
+            },
+        )
 
     def get_attribute_value(self, attribute, attribute_value):
         # check attribute_value contains only text
@@ -233,13 +245,17 @@ class LoginView(ProfileMixin, LogMixin, View):
                     if content is not None:
                         values.append(content)
         attributes['issuer'] = login.remoteProviderId
+        in_response_to = login.response.inResponseTo
+        if in_response_to:
+            attributes['nonce'] = request.session.get('mellon-nonce-%s' % in_response_to)
+            attributes['force_authn'] = request.session.get('mellon-force-authn-%s' % in_response_to, False)
+
         if login.nameIdentifier:
             name_id = login.nameIdentifier
             name_id_format = force_text(name_id.format or lasso.SAML2_NAME_IDENTIFIER_FORMAT_UNSPECIFIED)
-            attributes.update({
-                'name_id_content': lasso_decode(name_id.content),
-                'name_id_format': name_id_format
-            })
+            attributes.update(
+                {'name_id_content': lasso_decode(name_id.content), 'name_id_format': name_id_format}
+            )
             if name_id.nameQualifier:
                 attributes['name_id_name_qualifier'] = force_text(name_id.nameQualifier)
             if name_id.spNameQualifier:
@@ -249,65 +265,83 @@ class LoginView(ProfileMixin, LogMixin, View):
             attributes['authn_instant'] = utils.iso8601_to_datetime(authn_statement.authnInstant)
         if authn_statement.sessionNotOnOrAfter:
             attributes['session_not_on_or_after'] = utils.iso8601_to_datetime(
-                authn_statement.sessionNotOnOrAfter)
+                authn_statement.sessionNotOnOrAfter
+            )
         if authn_statement.sessionIndex:
             attributes['session_index'] = authn_statement.sessionIndex
         attributes['authn_context_class_ref'] = ()
         if authn_statement.authnContext:
             authn_context = authn_statement.authnContext
             if authn_context.authnContextClassRef:
-                attributes['authn_context_class_ref'] = \
-                    authn_context.authnContextClassRef
+                attributes['authn_context_class_ref'] = authn_context.authnContextClassRef
         self.log.debug('trying to authenticate with attributes %r', attributes)
         response = self.authenticate(request, login, attributes)
         response.delete_cookie(RETRY_LOGIN_COOKIE)
         return response
 
     def authenticate(self, request, login, attributes):
-        user = auth.authenticate(saml_attributes=attributes)
+        user = auth.authenticate(request=request, saml_attributes=attributes)
         next_url = self.get_next_url(default=resolve_url(settings.LOGIN_REDIRECT_URL))
         if user is not None:
             if user.is_active:
-                utils.login(request, user)
-                session_index = attributes['session_index']
-                if session_index:
-                    if not request.session.session_key:
-                        request.session.create()
-                    models.SessionIndex.objects.get_or_create(
-                        saml_identifier=user.saml_identifier,
-                        session_key=request.session.session_key,
-                        session_index=session_index)
-                self.log.info('user %s (NameID is %r) logged in using SAML', user,
-                              attributes['name_id_content'])
-                request.session['mellon_session'] = utils.flatten_datetime(attributes)
-                if ('session_not_on_or_after' in attributes and not settings.SESSION_EXPIRE_AT_BROWSER_CLOSE):
-                    request.session.set_expiry(
-                        utils.get_seconds_expiry(
-                            attributes['session_not_on_or_after']))
+                self.login(user, attributes)
             else:
-                self.log.warning('user %s (NameID is %r) is inactive, login refused', user,
-                                 attributes['name_id_content'])
-                return self.render(request, 'mellon/inactive_user.html', {
-                    'user': user,
-                    'saml_attributes': attributes})
+                self.log.warning(
+                    'user %s (NameID is %r) is inactive, login refused', user, attributes['name_id_content']
+                )
+                return self.render(
+                    request, 'mellon/inactive_user.html', {'user': user, 'saml_attributes': attributes}
+                )
         else:
             self.log.warning('no user found for NameID %r', attributes['name_id_content'])
-            return self.render(
-                request, 'mellon/user_not_found.html',
-                {'saml_attributes': attributes})
+            return self.render(request, 'mellon/user_not_found.html', {'saml_attributes': attributes})
         request.session['lasso_session_dump'] = login.session.dump()
 
-        return HttpResponseRedirect(next_url)
+        if self.debug_login:
+            return self.render_debug_template(request, login, attributes)
+        else:
+            return HttpResponseRedirect(next_url)
+
+    def render_debug_template(self, request, login, attributes):
+        request.session['mellon_debug_login'] = False
+        context = {
+            'logs': self.stream.getvalue(),
+            'attributes': attributes,
+            'login': login,
+            'response_dump': login.response and login.response.debug(4),
+            'assertion_dump': login.assertion and login.assertion.debug(4),
+        }
+        return self.render(request, 'mellon/debug_login.html', context)
+
+    def login(self, user, attributes):
+        if self.debug_login:
+            self.log.info('mellon: would login user %s (username %s)', user.get_full_name(), user)
+            return
+
+        utils.login(self.request, user)
+        session_index = attributes['session_index']
+        if session_index:
+            if not self.request.session.session_key:
+                self.request.session.create()
+            models.SessionIndex.objects.get_or_create(
+                saml_identifier=user.saml_identifier,
+                session_key=self.request.session.session_key,
+                session_index=session_index,
+            )
+        self.log.info('user %s (NameID is %r) logged in using SAML', user, attributes['name_id_content'])
+        self.request.session['mellon_session'] = utils.flatten_datetime(attributes)
+        if 'session_not_on_or_after' in attributes and not settings.SESSION_EXPIRE_AT_BROWSER_CLOSE:
+            self.request.session.set_expiry(utils.get_seconds_expiry(attributes['session_not_on_or_after']))
 
     def retry_login(self):
-        '''Retry login if it failed for a temporary error.
+        """Retry login if it failed for a temporary error.
 
-           Use a cookie to prevent looping forever.
-        '''
+        Use a cookie to prevent looping forever.
+        """
         if RETRY_LOGIN_COOKIE in self.request.COOKIES:
             response = self.sso_failure(
-                self.request,
-                reason=_('There were too many redirections with the identity provider.'))
+                self.request, reason=_('There were too many redirections with the identity provider.')
+            )
             response.delete_cookie(RETRY_LOGIN_COOKIE)
             return response
         url = reverse('mellon_login')
@@ -341,28 +375,33 @@ class LoginView(ProfileMixin, LogMixin, View):
             return HttpResponseBadRequest('artifact is malformed %r' % artifact)
         except lasso.ServerProviderNotFoundError:
             self.log.warning('no entity id found for artifact %s', artifact)
-            return HttpResponseBadRequest(
-                'no entity id found for this artifact %r' % artifact)
+            return HttpResponseBadRequest('no entity id found for this artifact %r' % artifact)
         idp = utils.get_idp(login.remoteProviderId)
         if not idp:
-            return HttpResponseBadRequest(
-                'entity id %r is unknown' % login.remoteProviderId)
-        verify_ssl_certificate = utils.get_setting(
-            idp, 'VERIFY_SSL_CERTIFICATE')
+            return HttpResponseBadRequest('entity id %r is unknown' % login.remoteProviderId)
+        verify_ssl_certificate = utils.get_setting(idp, 'VERIFY_SSL_CERTIFICATE')
         login.buildRequestMsg()
         try:
-            result = requests.post(login.msgUrl, data=login.msgBody,
-                                   headers={'content-type': 'text/xml'},
-                                   timeout=app_settings.ARTIFACT_RESOLVE_TIMEOUT,
-                                   verify=verify_ssl_certificate)
+            result = requests.post(
+                login.msgUrl,
+                data=login.msgBody,
+                headers={'content-type': 'text/xml'},
+                timeout=app_settings.ARTIFACT_RESOLVE_TIMEOUT,
+                verify=verify_ssl_certificate,
+            )
         except RequestException as e:
             self.log.warning('unable to reach %r: %s', login.msgUrl, e)
-            return self.sso_failure(request,
-                                    reason=_('IdP is temporarily down, please try again ' 'later.'),
-                                    status_codes=status_codes)
+            return self.sso_failure(
+                request,
+                reason=_('IdP is temporarily down, please try again ' 'later.'),
+                status_codes=status_codes,
+            )
         if result.status_code != 200:
-            self.log.warning('SAML authentication failed: IdP returned %s when given artifact: %r',
-                             result.status_code, result.content)
+            self.log.warning(
+                'SAML authentication failed: IdP returned %s when given artifact: %r',
+                result.status_code,
+                result.content,
+            )
             return self.sso_failure(request, reason=idp_message, status_codes=status_codes)
 
         self.log.info('Got SAML Artifact Response', extra={'saml_response': result.content})
@@ -377,18 +416,20 @@ class LoginView(ProfileMixin, LogMixin, View):
         except lasso.ProfileInvalidMsgError:
             self.log.warning('ArtifactResolveResponse is malformed %r', result.content[:200])
             if settings.DEBUG:
-                return HttpResponseBadRequest('ArtififactResolveResponse is malformed\n%r' %
-                                              result.content)
+                return HttpResponseBadRequest('ArtififactResolveResponse is malformed\n%r' % result.content)
             else:
                 return HttpResponseBadRequest('ArtififactResolveResponse is malformed')
         except lasso.ProfileCannotVerifySignatureError:
-            self.log.warning('SAML authentication failed: signature validation failed for %r',
-                             login.remoteProviderId)
+            self.log.warning(
+                'SAML authentication failed: signature validation failed for %r', login.remoteProviderId
+            )
         except lasso.ParamError:
             self.log.exception('lasso param error')
-        except (lasso.LoginStatusNotSuccessError,
-                lasso.ProfileStatusNotSuccessError,
-                lasso.ProfileRequestDeniedError):
+        except (
+            lasso.LoginStatusNotSuccessError,
+            lasso.ProfileStatusNotSuccessError,
+            lasso.ProfileRequestDeniedError,
+        ):
             status = login.response.status
             a = status
             while a.statusCode:
@@ -414,8 +455,7 @@ class LoginView(ProfileMixin, LogMixin, View):
         url = app_settings.DISCOVERY_SERVICE_URL
         params = {
             # prevent redirect loops with the discovery service
-            'entityID': request.build_absolute_uri(
-                reverse('mellon_metadata')),
+            'entityID': request.build_absolute_uri(reverse('mellon_metadata')),
             'return': return_url,
         }
         if is_passive:
@@ -426,14 +466,17 @@ class LoginView(ProfileMixin, LogMixin, View):
     def get(self, request, *args, **kwargs):
         '''Initialize login request'''
         if 'SAMLart' in request.GET:
+            if 'artifact' not in app_settings.ASSERTION_CONSUMER_BINDINGS:
+                raise Http404('artifact binding is not supported')
             return self.continue_sso_artifact(request, lasso.HTTP_METHOD_ARTIFACT_GET)
 
         # redirect to discovery service if needed
-        if ('entityID' not in request.GET
-                and 'nodisco' not in request.GET
-                and app_settings.DISCOVERY_SERVICE_URL):
-            return self.request_discovery_service(
-                request, is_passive=request.GET.get('passive') == '1')
+        if (
+            'entityID' not in request.GET
+            and 'nodisco' not in request.GET
+            and app_settings.DISCOVERY_SERVICE_URL
+        ):
+            return self.request_discovery_service(request, is_passive=request.GET.get('passive') == '1')
 
         next_url = check_next_url(self.request, request.GET.get(REDIRECT_FIELD_NAME))
         idp = self.get_idp(request)
@@ -448,15 +491,19 @@ class LoginView(ProfileMixin, LogMixin, View):
             policy = authn_request.nameIdPolicy
             policy.allowCreate = utils.get_setting(idp, 'NAME_ID_POLICY_ALLOW_CREATE')
             policy.format = utils.get_setting(idp, 'NAME_ID_POLICY_FORMAT')
-            force_authn = utils.get_setting(idp, 'FORCE_AUTHN')
+            force_authn = utils.get_setting(idp, 'FORCE_AUTHN') or 'force-authn' in request.GET
+            # link the nonce and forceAuthn state to the request-id
+            if 'nonce' in request.GET:
+                request.session['mellon-nonce-%s' % authn_request.id] = request.GET['nonce']
             if force_authn:
+                request.session['mellon-force-authn-%s' % authn_request.id] = True
                 authn_request.forceAuthn = True
             if request.GET.get('passive') == '1':
                 authn_request.isPassive = True
             # configure requested AuthnClassRef
             authn_classref = utils.get_setting(idp, 'AUTHN_CLASSREF')
             if authn_classref:
-                authn_classref = tuple([str(x) for x in authn_classref])
+                authn_classref = tuple(str(x) for x in authn_classref)
                 req_authncontext = lasso.Samlp2RequestedAuthnContext()
                 authn_request.requestedAuthnContext = req_authncontext
                 req_authncontext.authnContextClassRef = authn_classref
@@ -467,15 +514,21 @@ class LoginView(ProfileMixin, LogMixin, View):
                 # lasso>2.5.1 introduced a better API
                 if hasattr(authn_request.extensions, 'any'):
                     authn_request.extensions.any = (
-                        str('<eo:next_url xmlns:eo="https://www.entrouvert.com/">%s</eo:next_url>' % eo_next_url),
+                        str(
+                            '<eo:next_url xmlns:eo="https://www.entrouvert.com/">%s</eo:next_url>'
+                            % eo_next_url
+                        ),
                     )
                 else:
                     authn_request.extensions.setOriginalXmlnode(
-                        str('''<samlp:Extensions
+                        str(
+                            '''<samlp:Extensions
                                  xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                                  xmlns:eo="https://www.entrouvert.com/">
                                <eo:next_url>%s</eo:next_url>
-                            </samlp:Extensions>''' % eo_next_url)
+                            </samlp:Extensions>'''
+                            % eo_next_url
+                        )
                     )
             self.set_next_url(next_url)
             self.add_login_hints(idp, authn_request, request=request, next_url=next_url or '/')
@@ -493,8 +546,7 @@ class LoginView(ProfileMixin, LogMixin, View):
         assert hasattr(authn_request.extensions, 'any'), 'extension nodes need lasso > 2.5.1'
         serialized = ET.tostring(node, 'utf-8')
         # tostring return bytes in PY3, but lasso needs str
-        if six.PY3:
-            serialized = serialized.decode('utf-8')
+        serialized = serialized.decode('utf-8')
         extension_content = authn_request.extensions.any or ()
         extension_content += (serialized,)
         authn_request.extensions.any = extension_content
@@ -518,6 +570,17 @@ class LoginView(ProfileMixin, LogMixin, View):
             node.text = hint
             self.add_extension_node(authn_request, node)
 
+    @contextmanager
+    def capture_logs(self):
+        self.stream = StringIO()
+        handler = logging.StreamHandler(self.stream)
+        handler.setLevel(logging.DEBUG)
+        self.log.root.addHandler(handler)
+        try:
+            yield
+        finally:
+            self.log.root.removeHandler(handler)
+
 
 # we need fine control of transactions to prevent double user creations
 login = transaction.non_atomic_requests(csrf_exempt(LoginView.as_view()))
@@ -548,7 +611,9 @@ class LogoutView(ProfileMixin, LogMixin, View):
                 # there is not current session
                 not request.user.is_authenticated
                 # or the current session is not part of the list
-                or request.session.session_key not in session_keys))
+                or request.session.session_key not in session_keys
+            )
+        )
 
         if asynchronous_logout:
             current_session_key = request.session.session_key if request.user.is_authenticated else None
@@ -583,25 +648,26 @@ class LogoutView(ProfileMixin, LogMixin, View):
             return HttpResponseBadRequest('error processing logout request: %r' % e)
 
         issuer = force_text(logout.remoteProviderId)
-        session_indexes = set(force_text(sessionIndex) for sessionIndex in logout.request.sessionIndexes)
+        session_indexes = {force_text(sessionIndex) for sessionIndex in logout.request.sessionIndexes}
 
-        saml_identifier = models.UserSAMLIdentifier.objects.filter(
-            name_id=force_text(logout.nameIdentifier.content),
-            issuer=issuer).select_related('user').first()
+        saml_identifier = (
+            models.UserSAMLIdentifier.objects.filter(
+                name_id=force_text(logout.nameIdentifier.content), issuer=issuer
+            )
+            .select_related('user')
+            .first()
+        )
 
         if saml_identifier:
             name_id_user = saml_identifier.user
-            indexes = models.SessionIndex.objects.select_related(
-                'saml_identifier').filter(
-                    saml_identifier=saml_identifier)
+            indexes = models.SessionIndex.objects.select_related('saml_identifier').filter(
+                saml_identifier=saml_identifier
+            )
             if session_indexes:
                 indexes = indexes.filter(session_index__in=session_indexes)
 
             # lasso has too much state :/
-            logout.setSessionFromDump(
-                utils.make_session_dump(
-                    logout.nameIdentifier,
-                    indexes))
+            logout.setSessionFromDump(utils.make_session_dump(logout.nameIdentifier, indexes))
 
             try:
                 logout.validateRequest()
@@ -618,7 +684,8 @@ class LogoutView(ProfileMixin, LogMixin, View):
                     saml_user=name_id_user,
                     session_indexes=session_indexes,
                     indexes=indexes,
-                    mode=mode)
+                    mode=mode,
+                )
 
         try:
             logout.buildResponseMsg()
@@ -692,3 +759,17 @@ logout = csrf_exempt(LogoutView.as_view())
 def metadata(request, **kwargs):
     metadata = utils.create_metadata(request)
     return HttpResponse(metadata, content_type='text/xml')
+
+
+class DebugLoginView(RedirectView):
+    pattern_name = 'mellon_login'
+    query_string = True
+
+    def dispatch(self, request, *args, **kwargs):
+        if not settings.DEBUG:
+            return HttpResponseForbidden()
+        request.session['mellon_debug_login'] = True
+        return super().dispatch(request, *args, **kwargs)
+
+
+debug_login = csrf_exempt(DebugLoginView.as_view())
